local __exports = LibStub:NewLibrary("ovale/states/conditions", 90032)
if not __exports then return end
local __class = LibStub:GetLibrary("tslib").newClass
local LibBabbleCreatureType = LibStub:GetLibrary("LibBabble-CreatureType-3.0", true)
local LibRangeCheck = LibStub:GetLibrary("LibRangeCheck-2.0", true)
local __enginecondition = LibStub:GetLibrary("ovale/engine/condition")
local ParseCondition = __enginecondition.ParseCondition
local ReturnBoolean = __enginecondition.ReturnBoolean
local ReturnConstant = __enginecondition.ReturnConstant
local ReturnValue = __enginecondition.ReturnValue
local ReturnValueBetween = __enginecondition.ReturnValueBetween
local ipairs = ipairs
local pairs = pairs
local type = type
local GetBuildInfo = GetBuildInfo
local GetItemCount = GetItemCount
local GetNumTrackingTypes = GetNumTrackingTypes
local GetTrackingInfo = GetTrackingInfo
local GetUnitSpeed = GetUnitSpeed
local HasFullControl = HasFullControl
local IsStealthed = IsStealthed
local UnitCastingInfo = UnitCastingInfo
local UnitChannelInfo = UnitChannelInfo
local UnitClass = UnitClass
local UnitClassification = UnitClassification
local UnitCreatureFamily = UnitCreatureFamily
local UnitCreatureType = UnitCreatureType
local UnitDetailedThreatSituation = UnitDetailedThreatSituation
local UnitExists = UnitExists
local UnitInParty = UnitInParty
local UnitInRaid = UnitInRaid
local UnitIsDead = UnitIsDead
local UnitIsFriend = UnitIsFriend
local UnitIsPVP = UnitIsPVP
local UnitIsUnit = UnitIsUnit
local UnitLevel = UnitLevel
local UnitName = UnitName
local UnitPower = UnitPower
local UnitPowerMax = UnitPowerMax
local UnitRace = UnitRace
local INFINITY = math.huge
local min = math.min
local lower = string.lower
local upper = string.upper
local sub = string.sub
local __toolstools = LibStub:GetLibrary("ovale/tools/tools")
local isNumber = __toolstools.isNumber
local OneTimeMessage = __toolstools.OneTimeMessage
local function BossArmorDamageReduction(target)
    return 0.3
end
local function Capitalize(word)
    if  not word then
        return word
    end
    return upper(sub(word, 1, 1)) .. lower(sub(word, 2))
end
local AMPLIFICATION = 146051
local INCREASED_CRIT_EFFECT_3_PERCENT = 44797
local IMBUED_BUFF_ID = 214336
local STEADY_FOCUS = 177668
local checkHaste = {
    base = true,
    melee = true,
    none = true,
    ranged = true,
    spell = true
}
local mapFilter = {
    buff = "HELPFUL",
    debuff = "HARMFUL"
}
__exports.OvaleConditions = __class(nil, {
    ComputeParameter = function(self, spellId, paramName, atTime)
        return self.OvaleData:GetSpellInfoProperty(spellId, atTime, paramName, nil)
    end,
    GetHastedTime = function(self, seconds, haste)
        seconds = seconds or 0
        local multiplier = self.OvalePaperDoll:GetHasteMultiplier(haste, self.OvalePaperDoll.next)
        return seconds / multiplier
    end,
    GetDiseases = function(self, target, atTime)
        local bpAura = self.OvaleAura:GetAura(target, 55078, atTime, "HARMFUL", true)
        local ffAura = self.OvaleAura:GetAura(target, 55095, atTime, "HARMFUL", true)
        return bpAura, ffAura
    end,
    MaxPower = function(self, powerType, positionalParams, namedParams, atTime)
        local target = self:ParseCondition(positionalParams, namedParams)
        local value
        if target == "player" then
            value = self.OvalePower.current.maxPower[powerType]
        else
            local powerInfo = self.OvalePower.POWER_INFO[powerType]
            value = (powerInfo and UnitPowerMax(target, powerInfo.id, powerInfo.segments)) or 0
        end
        return ReturnConstant(value)
    end,
    Power = function(self, powerType, positionalParams, namedParams, atTime)
        local target = self:ParseCondition(positionalParams, namedParams)
        if target == "player" then
            local value = self.OvalePower.next.power[powerType] or 0
            local rate = self.OvalePower:getPowerRateAt(self.OvalePower.next, powerType, atTime)
            return ReturnValueBetween(atTime, INFINITY, value, atTime, rate)
        else
            local powerInfo = self.OvalePower.POWER_INFO[powerType]
            local value = (powerInfo and UnitPower(target, powerInfo.id)) or 0
            return ReturnConstant(value)
        end
    end,
    PowerDeficit = function(self, powerType, positionalParams, namedParams, atTime)
        local target = self:ParseCondition(positionalParams, namedParams)
        if target == "player" then
            local powerMax = self.OvalePower.current.maxPower[powerType] or 0
            if powerMax > 0 then
                local power = self.OvalePower.next.power[powerType] or 0
                local rate = self.OvalePower:getPowerRateAt(self.OvalePower.next, powerType, atTime)
                return ReturnValueBetween(atTime, INFINITY, powerMax - power, atTime, -1 * rate)
            end
        else
            local powerInfo = self.OvalePower.POWER_INFO[powerType]
            local powerMax = (powerInfo and UnitPowerMax(target, powerInfo.id, powerInfo.segments)) or 0
            if powerMax > 0 then
                local power = (powerInfo and UnitPower(target, powerInfo.id)) or 0
                local value = powerMax - power
                return ReturnConstant(value)
            end
        end
        return ReturnConstant(0)
    end,
    PowerPercent = function(self, powerType, positionalParams, namedParams, atTime)
        local target = self:ParseCondition(positionalParams, namedParams)
        if target == "player" then
            local powerMax = self.OvalePower.current.maxPower[powerType] or 0
            if powerMax > 0 then
                local ratio = 100 / powerMax
                local power = self.OvalePower.next.power[powerType] or 0
                local rate = ratio * self.OvalePower:getPowerRateAt(self.OvalePower.next, powerType, atTime)
                if (rate > 0 and power >= powerMax) or (rate < 0 and power == 0) then
                    rate = 0
                end
                return ReturnValueBetween(atTime, INFINITY, power * ratio, atTime, rate)
            end
        else
            local powerInfo = self.OvalePower.POWER_INFO[powerType]
            local powerMax = (powerInfo and UnitPowerMax(target, powerInfo.id, powerInfo.segments)) or 0
            if powerMax > 0 then
                local ratio = 100 / powerMax
                local value = (powerInfo and ratio * UnitPower(target, powerInfo.id)) or 0
                return ReturnConstant(value)
            end
        end
        return ReturnConstant(0)
    end,
    PowerCost = function(self, powerType, positionalParams, namedParams, atTime)
        local spell = positionalParams[1]
        local spellId = self.OvaleSpellBook:getKnownSpellId(spell)
        local target = self:ParseCondition(positionalParams, namedParams, "target")
        local maxCost = namedParams.max == 1
        local value = (spellId and self.OvalePower:PowerCost(spellId, powerType, atTime, target, maxCost)) or 0
        return ReturnConstant(value)
    end,
    Snapshot = function(self, statName, defaultValue, positionalParams, namedParams, atTime)
        local value = self.OvalePaperDoll:GetState(atTime)[statName] or defaultValue
        return ReturnConstant(value)
    end,
    SnapshotCritChance = function(self, statName, defaultValue, positionalParams, namedParams, atTime)
        local value = self.OvalePaperDoll:GetState(atTime)[statName] or defaultValue
        if namedParams.unlimited ~= 1 and value > 100 then
            value = 100
        end
        return ReturnConstant(value)
    end,
    TimeToPower = function(self, powerType, level, atTime)
        level = level or 0
        local seconds = self.OvalePower:getTimeToPowerAt(self.OvalePower.next, level, powerType, atTime)
        if seconds == 0 then
            return ReturnConstant(0)
        elseif seconds < INFINITY then
            return ReturnValueBetween(0, atTime + seconds, seconds, atTime, -1)
        else
            return ReturnConstant(INFINITY)
        end
    end,
    TimeToPowerFor = function(self, powerType, positionalParams, namedParams, atTime)
        local spellId = positionalParams[1]
        local target = self:ParseCondition(positionalParams, namedParams, "target")
        local targetGuid = self.OvaleGUID:UnitGUID(target)
        if  not targetGuid then
            return 
        end
        local seconds = self.OvaleSpells:TimeToPowerForSpell(spellId, atTime, targetGuid, powerType)
        if seconds == 0 then
            return ReturnConstant(0)
        elseif seconds < INFINITY then
            return ReturnValueBetween(0, atTime + seconds, seconds, atTime, -1)
        else
            return ReturnConstant(INFINITY)
        end
    end,
    ParseCondition = function(self, positionalParams, namedParams, defaultTarget)
        return ParseCondition(namedParams, self.baseState, defaultTarget)
    end,
    constructor = function(self, ovaleCondition, OvaleData, OvalePaperDoll, OvaleAzeriteEssence, OvaleAura, baseState, OvaleCooldown, OvaleFuture, OvaleSpellBook, OvaleFrameModule, OvaleGUID, OvaleDamageTaken, OvalePower, OvaleEnemies, lastSpell, OvaleHealth, ovaleOptions, OvaleLossOfControl, OvaleSpellDamage, OvaleTotem, OvaleSigil, OvaleDemonHunterSoulFragments, OvaleRunes, OvaleBossMod, OvaleSpells)
        self.OvaleData = OvaleData
        self.OvalePaperDoll = OvalePaperDoll
        self.OvaleAzeriteEssence = OvaleAzeriteEssence
        self.OvaleAura = OvaleAura
        self.baseState = baseState
        self.OvaleCooldown = OvaleCooldown
        self.OvaleFuture = OvaleFuture
        self.OvaleSpellBook = OvaleSpellBook
        self.OvaleFrameModule = OvaleFrameModule
        self.OvaleGUID = OvaleGUID
        self.OvaleDamageTaken = OvaleDamageTaken
        self.OvalePower = OvalePower
        self.OvaleEnemies = OvaleEnemies
        self.lastSpell = lastSpell
        self.OvaleHealth = OvaleHealth
        self.ovaleOptions = ovaleOptions
        self.OvaleLossOfControl = OvaleLossOfControl
        self.OvaleSpellDamage = OvaleSpellDamage
        self.OvaleTotem = OvaleTotem
        self.OvaleSigil = OvaleSigil
        self.OvaleDemonHunterSoulFragments = OvaleDemonHunterSoulFragments
        self.OvaleRunes = OvaleRunes
        self.OvaleBossMod = OvaleBossMod
        self.OvaleSpells = OvaleSpells
        self.ArmorSetBonus = function(positionalParams, namedParams, atTime)
            OneTimeMessage("Warning: 'ArmorSetBonus()' is depreciated.  Returns 0")
            local value = 0
            return 0, INFINITY, value, 0, 0
        end
        self.ArmorSetParts = function(positionalParams, namedParams, atTime)
            local value = 0
            OneTimeMessage("Warning: 'ArmorSetBonus()' is depreciated.  Returns 0")
            return ReturnConstant(value)
        end
        self.AzeriteEssenceIsMajor = function(positionalParams, namedParams, atTime)
            local essenceId = positionalParams[1]
            local value = self.OvaleAzeriteEssence:IsMajorEssence(essenceId)
            return ReturnBoolean(value)
        end
        self.AzeriteEssenceIsMinor = function(positionalParams, namedParams, atTime)
            local essenceId = positionalParams[1]
            local value = self.OvaleAzeriteEssence:IsMinorEssence(essenceId)
            return ReturnBoolean(value)
        end
        self.AzeriteEssenceIsEnabled = function(positionalParams, namedParams, atTime)
            local essenceId = positionalParams[1]
            local value = self.OvaleAzeriteEssence:IsMajorEssence(essenceId) or self.OvaleAzeriteEssence:IsMinorEssence(essenceId)
            return ReturnBoolean(value)
        end
        self.AzeriteEssenceRank = function(positionalParams, namedParams, atTime)
            local essenceId = positionalParams[1]
            local value = self.OvaleAzeriteEssence:EssenceRank(essenceId)
            return ReturnConstant(value)
        end
        self.BaseDuration = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local value = 0
            if self.OvaleData.buffSpellList[auraId] then
                local spellList = self.OvaleData.buffSpellList[auraId]
                for id in pairs(spellList) do
                    value = self.OvaleAura:GetBaseDuration(id, self.OvalePaperDoll.next)
                    if value ~= INFINITY then
                        break
                    end
                end
            else
                value = self.OvaleAura:GetBaseDuration(auraId, self.OvalePaperDoll.next)
            end
            return ReturnConstant(value)
        end
        self.BuffAmount = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local value = namedParams.value or 1
            local statName = "value1"
            if value == 1 then
                statName = "value1"
            elseif value == 2 then
                statName = "value2"
            elseif value == 3 then
                statName = "value3"
            end
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                local value = aura[statName] or 0
                return ReturnValueBetween(aura.gain, aura.ending, value, aura.start, 0)
            end
            return ReturnConstant(0)
        end
        self.BuffComboPoints = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                local value = (aura and aura.combopoints) or 0
                return ReturnValueBetween(aura.gain, aura.ending, value, aura.start, 0)
            end
            return ReturnConstant(0)
        end
        self.BuffCooldown = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            if  not isNumber(auraId) then
                return 
            end
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura then
                local gain = aura.gain
                local cooldownEnding = aura.cooldownEnding or 0
                return ReturnValueBetween(gain, INFINITY, 0, cooldownEnding, -1)
            end
            return ReturnConstant(0)
        end
        self.BuffCount = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local spellList = self.OvaleData.buffSpellList[auraId]
            local count = 0
            for id in pairs(spellList) do
                local aura = self.OvaleAura:GetAura(target, id, atTime, filter, mine)
                if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                    count = count + 1
                end
            end
            return ReturnConstant(count)
        end
        self.BuffCooldownDuration = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local minCooldown = INFINITY
            if self.OvaleData.buffSpellList[auraId] then
                for id in pairs(self.OvaleData.buffSpellList[auraId]) do
                    local si = self.OvaleData.spellInfo[id]
                    local cd = si and si.buff_cd
                    if cd and minCooldown > cd then
                        minCooldown = cd
                    end
                end
            else
                minCooldown = 0
            end
            return ReturnConstant(minCooldown)
        end
        self.BuffCountOnAny = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local _, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local excludeUnitId = (namedParams.excludeTarget == 1 and self.baseState.defaultTarget) or nil
            local fractional = (namedParams.count == 0 and true) or false
            local count, _, startChangeCount, endingChangeCount, startFirst, endingLast = self.OvaleAura:AuraCount(auraId, filter, mine, namedParams.stacks, atTime, excludeUnitId)
            if count > 0 and startChangeCount < INFINITY and fractional then
                local rate = -1 / (endingChangeCount - startChangeCount)
                return ReturnValueBetween(startFirst, endingLast, count, startChangeCount, rate)
            end
            return ReturnConstant(count)
        end
        self.BuffDirection = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura then
                return ReturnValueBetween(aura.gain, INFINITY, aura.direction, aura.gain, 0)
            end
            return ReturnConstant(0)
        end
        self.BuffDuration = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                local duration = aura.ending - aura.start
                return ReturnValueBetween(aura.gain, aura.ending, duration, aura.start, 0)
            end
            return ReturnConstant(0)
        end
        self.BuffExpires = function(positionalParams, namedParams, atTime)
            local auraId, seconds = positionalParams[1], positionalParams[2] or 0
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            if  not isNumber(auraId) or  not isNumber(seconds) then
                return 
            end
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura then
                local gain, _, ending = aura.gain, aura.start, aura.ending
                local hastedSeconds = self:GetHastedTime(seconds, namedParams.haste)
                if ending - hastedSeconds <= gain then
                    return gain, INFINITY
                else
                    return ending - hastedSeconds, INFINITY
                end
            end
            return 0, INFINITY
        end
        self.BuffPresent = function(atTime, auraId, target, filter, mine, seconds, haste)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura then
                local gain, _, ending = aura.gain, aura.start, aura.ending
                seconds = self:GetHastedTime(seconds, haste)
                if ending - seconds <= gain then
                    return 
                else
                    return gain, ending - seconds
                end
            end
            return 
        end
        self.BuffGain = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura then
                local gain = aura.gain or 0
                return ReturnValueBetween(gain, INFINITY, 0, gain, 1)
            end
            return ReturnConstant(0)
        end
        self.BuffImproved = function(positionalParams, namedParams, atTime)
            local _, _ = self:ParseCondition(positionalParams, namedParams)
            return ReturnConstant(0)
        end
        self.BuffPersistentMultiplier = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                local value = aura.damageMultiplier or 1
                return ReturnValueBetween(aura.gain, aura.ending, value, aura.start, 0)
            end
            return ReturnConstant(1)
        end
        self.BuffRemaining = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura and aura.ending >= atTime then
                return ReturnValueBetween(aura.gain, INFINITY, 0, aura.ending, -1)
            end
            return ReturnConstant(0)
        end
        self.BuffRemainingOnAny = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local _, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local excludeUnitId = (namedParams.excludeTarget == 1 and self.baseState.defaultTarget) or nil
            local count, _, _, _, startFirst, endingLast = self.OvaleAura:AuraCount(auraId, filter, mine, namedParams.stacks, atTime, excludeUnitId)
            if count > 0 then
                return ReturnValueBetween(startFirst, INFINITY, 0, endingLast, -1)
            end
            return ReturnConstant(0)
        end
        self.BuffStacks = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                local value = aura.stacks or 0
                return ReturnValueBetween(aura.gain, aura.ending, value, aura.start, 0)
            end
            return ReturnConstant(0)
        end
        self.maxStacks = function(positionalParams, namedParameters, atTime)
            local auraId = positionalParams[1]
            local spellInfo = self.OvaleData:GetSpellInfo(auraId)
            local maxStacks = (spellInfo and spellInfo.max_stacks) or 0
            return ReturnConstant(maxStacks)
        end
        self.BuffStacksOnAny = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local _, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local excludeUnitId = (namedParams.excludeTarget == 1 and self.baseState.defaultTarget) or nil
            local count, stacks, _, endingChangeCount, startFirst = self.OvaleAura:AuraCount(auraId, filter, mine, 1, atTime, excludeUnitId)
            if count > 0 then
                return ReturnValueBetween(startFirst, endingChangeCount, stacks, startFirst, 0)
            end
            return ReturnConstant(count)
        end
        self.BuffStealable = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            return self.OvaleAura:GetAuraWithProperty(target, "stealable", "HELPFUL", atTime)
        end
        self.CanCast = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local start, duration = self.OvaleCooldown:GetSpellCooldown(spellId, atTime)
            return start + duration, INFINITY
        end
        self.CastTime = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local castTime = self.OvaleSpellBook:GetCastTime(spellId) or 0
            return ReturnConstant(castTime)
        end
        self.ExecuteTime = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local castTime = self.OvaleSpellBook:GetCastTime(spellId) or 0
            local gcd = self.OvaleFuture:GetGCD(atTime)
            local t = (castTime > gcd and castTime) or gcd
            return ReturnConstant(t)
        end
        self.Casting = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams)
            local start, ending, castSpellId, castSpellName
            if target == "player" then
                start = self.OvaleFuture.next.currentCast.start
                ending = self.OvaleFuture.next.currentCast.stop
                castSpellId = self.OvaleFuture.next.currentCast.spellId
                castSpellName = self.OvaleSpellBook:GetSpellName(castSpellId)
            else
                local spellName, _, _, startTime, endTime = UnitCastingInfo(target)
                if  not spellName then
                    spellName, _, _, startTime, endTime = UnitChannelInfo(target)
                end
                if spellName then
                    castSpellName = spellName
                    start = startTime / 1000
                    ending = endTime / 1000
                end
            end
            if (castSpellId or castSpellName) and start and ending then
                if  not spellId then
                    return start, ending
                elseif self.OvaleData.buffSpellList[spellId] then
                    for id in pairs(self.OvaleData.buffSpellList[spellId]) do
                        if id == castSpellId or self.OvaleSpellBook:GetSpellName(id) == castSpellName then
                            return start, ending
                        end
                    end
                elseif spellId == "harmful" and self.OvaleSpellBook:IsHarmfulSpell(spellId) then
                    return start, ending
                elseif spellId == "helpful" and self.OvaleSpellBook:IsHelpfulSpell(spellId) then
                    return start, ending
                elseif spellId == castSpellId then
                    OneTimeMessage("%f %f %d %s => %d (%f)", start, ending, castSpellId, castSpellName, spellId, self.baseState.currentTime)
                    return start, ending
                elseif type(spellId) == "number" and self.OvaleSpellBook:GetSpellName(spellId) == castSpellName then
                    return start, ending
                end
            end
            return 
        end
        self.CheckBoxOff = function(positionalParams, namedParams, atTime)
            for _, id in ipairs(positionalParams) do
                if self.OvaleFrameModule.frame and self.OvaleFrameModule.frame:IsChecked(id) then
                    return 
                end
            end
            return 0, INFINITY
        end
        self.CheckBoxOn = function(positionalParams, namedParams, atTime)
            for _, id in ipairs(positionalParams) do
                if self.OvaleFrameModule.frame and  not self.OvaleFrameModule.frame:IsChecked(id) then
                    return 
                end
            end
            return 0, INFINITY
        end
        self.Class = function(positionalParams, namedParams, atTime)
            local className = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams)
            local classToken
            if target == "player" then
                classToken = self.OvalePaperDoll.class
            else
                _, classToken = UnitClass(target)
            end
            local boolean = classToken == upper(className)
            return ReturnBoolean(boolean)
        end
        self.Classification = function(positionalParams, namedParams, atTime)
            local classification = positionalParams[1]
            local targetClassification
            local target = self:ParseCondition(positionalParams, namedParams)
            if UnitLevel(target) < 0 then
                targetClassification = "worldboss"
            elseif UnitExists("boss1") and self.OvaleGUID:UnitGUID(target) == self.OvaleGUID:UnitGUID("boss1") then
                targetClassification = "worldboss"
            else
                local aura = self.OvaleAura:GetAura(target, IMBUED_BUFF_ID, atTime, "HARMFUL", false)
                if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                    targetClassification = "worldboss"
                else
                    targetClassification = UnitClassification(target)
                    if targetClassification == "rareelite" then
                        targetClassification = "elite"
                    elseif targetClassification == "rare" then
                        targetClassification = "normal"
                    end
                end
            end
            local boolean = targetClassification == classification
            return ReturnBoolean(boolean)
        end
        self.Counter = function(positionalParams, namedParams, atTime)
            local counter = positionalParams[1]
            local value = self.OvaleFuture:GetCounter(counter, atTime)
            return ReturnConstant(value)
        end
        self.CreatureFamily = function(_, name, target)
            name = Capitalize(name)
            local family = UnitCreatureFamily(target)
            local lookupTable = LibBabbleCreatureType and LibBabbleCreatureType:GetLookupTable()
            return ReturnBoolean(lookupTable and family == lookupTable[name])
        end
        self.CreatureType = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local creatureType = UnitCreatureType(target)
            local lookupTable = LibBabbleCreatureType and LibBabbleCreatureType:GetLookupTable()
            if lookupTable then
                for _, name in ipairs(positionalParams) do
                    local capitalizedName = Capitalize(name)
                    if creatureType == lookupTable[capitalizedName] then
                        return 0, INFINITY
                    end
                end
            end
            return 
        end
        self.CritDamage = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            local value = self:ComputeParameter(spellId, "damage", atTime) or 0
            local si = self.OvaleData.spellInfo[spellId]
            if si and si.physical == 1 then
                value = value * (1 - BossArmorDamageReduction(target))
            end
            local critMultiplier = 2
            do
                local aura = self.OvaleAura:GetAura("player", AMPLIFICATION, atTime, "HELPFUL")
                if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                    critMultiplier = critMultiplier + (aura.value1 or 0)
                end
            end
            do
                local aura = self.OvaleAura:GetAura("player", INCREASED_CRIT_EFFECT_3_PERCENT, atTime, "HELPFUL")
                if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                    critMultiplier = critMultiplier * (aura.value1 or 0)
                end
            end
            value = critMultiplier * value
            return ReturnConstant(value)
        end
        self.Damage = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            local value = self:ComputeParameter(spellId, "damage", atTime) or 0
            local si = self.OvaleData.spellInfo[spellId]
            if si and si.physical == 1 then
                value = value * (1 - BossArmorDamageReduction(target))
            end
            return ReturnConstant(value)
        end
        self.DamageTaken = function(positionalParams, namedParams, atTime)
            local interval = positionalParams[1]
            local value = 0
            if interval > 0 then
                local total = self.OvaleDamageTaken:GetRecentDamage(interval)
                value = total
            end
            return ReturnConstant(value)
        end
        self.MagicDamageTaken = function(positionalParams, namedParams, atTime)
            local interval = positionalParams[1]
            local value = 0
            if interval > 0 then
                local _, totalMagic = self.OvaleDamageTaken:GetRecentDamage(interval)
                value = totalMagic
            end
            return ReturnConstant(value)
        end
        self.PhysicalDamageTaken = function(positionalParams, namedParams, atTime)
            local interval = positionalParams[1]
            local value = 0
            if interval > 0 then
                local total, totalMagic = self.OvaleDamageTaken:GetRecentDamage(interval)
                value = total - totalMagic
            end
            return ReturnConstant(value)
        end
        self.DiseasesRemaining = function(positionalParams, namedParams, atTime)
            local target, _ = self:ParseCondition(positionalParams, namedParams)
            local bpAura, ffAura = self:GetDiseases(target, atTime)
            local aura
            if bpAura and self.OvaleAura:IsActiveAura(bpAura, atTime) and ffAura and self.OvaleAura:IsActiveAura(ffAura, atTime) then
                aura = (bpAura.ending < ffAura.ending and bpAura) or ffAura
            end
            if aura then
                return ReturnValueBetween(aura.gain, INFINITY, 0, aura.ending, -1)
            end
            return ReturnConstant(0)
        end
        self.DiseasesTicking = function(positionalParams, namedParams, atTime)
            local target, _ = self:ParseCondition(positionalParams, namedParams)
            local bpAura, ffAura = self:GetDiseases(target, atTime)
            local gain, ending
            if bpAura and ffAura then
                gain = (bpAura.gain > ffAura.gain and bpAura.gain) or ffAura.gain
                ending = (bpAura.ending < ffAura.ending and bpAura.ending) or ffAura.ending
            end
            if gain and ending and ending > gain then
                return gain, ending
            end
            return 
        end
        self.DiseasesAnyTicking = function(positionalParams, namedParams, atTime)
            local target, _ = self:ParseCondition(positionalParams, namedParams)
            local bpAura, ffAura = self:GetDiseases(target, atTime)
            local aura
            if bpAura or ffAura then
                aura = bpAura or ffAura
                if bpAura and ffAura then
                    aura = (bpAura.ending > ffAura.ending and bpAura) or ffAura
                end
            end
            if aura then
                local gain, _, ending = aura.gain, aura.start, aura.ending
                if ending > gain then
                    return gain, ending
                end
            end
            return 
        end
        self.Distance = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local value = (LibRangeCheck and LibRangeCheck:GetRange(target)) or 0
            return ReturnConstant(value)
        end
        self.Enemies = function(positionalParams, namedParams, atTime)
            local value = self.OvaleEnemies.next.enemies
            if  not value then
                local useTagged = self.ovaleOptions.db.profile.apparence.taggedEnemies
                if namedParams.tagged == 0 then
                    useTagged = false
                elseif namedParams.tagged == 1 then
                    useTagged = true
                end
                value = (useTagged and self.OvaleEnemies.next.taggedEnemies) or self.OvaleEnemies.next.activeEnemies
            end
            if value < 1 then
                value = 1
            end
            return ReturnConstant(value)
        end
        self.EnergyRegenRate = function(positionalParams, namedParams, atTime)
            local value = self.OvalePower:getPowerRateAt(self.OvalePower.next, "energy", atTime)
            return ReturnConstant(value)
        end
        self.EnrageRemaining = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, "enrage", atTime, "HELPFUL", false)
            if aura and aura.ending >= atTime then
                return ReturnValueBetween(aura.gain, INFINITY, 0, aura.ending, -1)
            end
            return ReturnConstant(0)
        end
        self.Exists = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local boolean = UnitExists(target)
            return ReturnBoolean(boolean)
        end
        self.False = function(positionalParams, namedParams, atTime)
            return 
        end
        self.FocusRegenRate = function(positionalParams, namedParams, atTime)
            local value = self.OvalePower:getPowerRateAt(self.OvalePower.next, "focus", atTime)
            return ReturnConstant(value)
        end
        self.FocusCastingRegen = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local regenRate = self.OvalePower:getPowerRateAt(self.OvalePower.next, "focus", atTime)
            local power = 0
            local castTime = self.OvaleSpellBook:GetCastTime(spellId) or 0
            local gcd = self.OvaleFuture:GetGCD(atTime)
            local castSeconds = (castTime > gcd and castTime) or gcd
            power = power + regenRate * castSeconds
            local aura = self.OvaleAura:GetAura("player", STEADY_FOCUS, atTime, "HELPFUL", true)
            if aura then
                local seconds = aura.ending - atTime
                if seconds <= 0 then
                    seconds = 0
                elseif seconds > castSeconds then
                    seconds = castSeconds
                end
                power = power + regenRate * 1.5 * seconds
            end
            return ReturnConstant(power)
        end
        self.GCD = function(positionalParams, namedParams, atTime)
            local value = self.OvaleFuture:GetGCD(atTime)
            return ReturnConstant(value)
        end
        self.GCDRemaining = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            if self.OvaleFuture.next.lastGCDSpellId then
                local duration = self.OvaleFuture:GetGCD(self.OvaleFuture.next.lastGCDSpellId, atTime, self.OvaleGUID:UnitGUID(target))
                local spellcast = self.lastSpell:LastInFlightSpell()
                local start = (spellcast and spellcast.start) or 0
                local ending = start + duration
                if atTime < ending then
                    return ReturnValueBetween(start, INFINITY, 0, ending, -1)
                end
            end
            return ReturnConstant(0)
        end
        self.Glyph = function(positionalParams, namedParams, atTime)
            return ReturnBoolean(false)
        end
        self.HasFullControlCondition = function(positionalParams, namedParams, atTime)
            local boolean = HasFullControl()
            return ReturnBoolean(boolean)
        end
        self.Health = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local health = self.OvaleHealth:UnitHealth(target) or 0
            if health > 0 then
                local now = self.baseState.currentTime
                local timeToDie = self.OvaleHealth:UnitTimeToDie(target)
                local rate = (-1 * health) / timeToDie
                return ReturnValueBetween(now, INFINITY, health, now, rate)
            end
            return ReturnConstant(0)
        end
        self.EffectiveHealth = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local health = self.OvaleHealth:UnitHealth(target) + self.OvaleHealth:UnitAbsorb(target) - self.OvaleHealth:UnitHealAbsorb(target) or 0
            local now = self.baseState.currentTime
            local timeToDie = self.OvaleHealth:UnitTimeToDie(target)
            local rate = (-1 * health) / timeToDie
            return ReturnValueBetween(now, INFINITY, health, now, rate)
        end
        self.HealthMissing = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local health = self.OvaleHealth:UnitHealth(target) or 0
            local maxHealth = self.OvaleHealth:UnitHealthMax(target) or 1
            if health > 0 then
                local now = self.baseState.currentTime
                local missing = maxHealth - health
                local timeToDie = self.OvaleHealth:UnitTimeToDie(target)
                local rate = health / timeToDie
                return ReturnValueBetween(now, INFINITY, missing, now, rate)
            end
            return ReturnConstant(maxHealth)
        end
        self.HealthPercent = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local health = self.OvaleHealth:UnitHealth(target) or 0
            if health > 0 then
                local now = self.baseState.currentTime
                local maxHealth = self.OvaleHealth:UnitHealthMax(target) or 1
                local healthPct = (health / maxHealth) * 100
                local timeToDie = self.OvaleHealth:UnitTimeToDie(target)
                local rate = (-1 * healthPct) / timeToDie
                return ReturnValueBetween(now, INFINITY, healthPct, now, rate)
            end
            return ReturnConstant(0)
        end
        self.EffectiveHealthPercent = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local health = self.OvaleHealth:UnitHealth(target) + self.OvaleHealth:UnitAbsorb(target) - self.OvaleHealth:UnitHealAbsorb(target) or 0
            local now = self.baseState.currentTime
            local maxHealth = self.OvaleHealth:UnitHealthMax(target) or 1
            local healthPct = (health / maxHealth) * 100
            local timeToDie = self.OvaleHealth:UnitTimeToDie(target)
            local rate = (-1 * healthPct) / timeToDie
            return ReturnValueBetween(now, INFINITY, healthPct, now, rate)
        end
        self.MaxHealth = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local value = self.OvaleHealth:UnitHealthMax(target)
            return ReturnConstant(value)
        end
        self.TimeToDie = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local now = self.baseState.currentTime
            local timeToDie = self.OvaleHealth:UnitTimeToDie(target)
            return ReturnValueBetween(now, INFINITY, timeToDie, now, -1)
        end
        self.TimeToHealthPercent = function(positionalParams, namedParams, atTime)
            local percent = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams)
            local health = self.OvaleHealth:UnitHealth(target) or 0
            if health > 0 then
                local maxHealth = self.OvaleHealth:UnitHealthMax(target) or 1
                local healthPct = (health / maxHealth) * 100
                if healthPct >= percent then
                    local now = self.baseState.currentTime
                    local timeToDie = self.OvaleHealth:UnitTimeToDie(target)
                    local t = (timeToDie * (healthPct - percent)) / healthPct
                    return ReturnValueBetween(now, now + t, t, now, -1)
                end
            end
            return ReturnConstant(0)
        end
        self.InFlightToTarget = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local boolean = self.OvaleFuture.next.currentCast.spellId == spellId or self.OvaleFuture:InFlight(spellId)
            return ReturnBoolean(boolean)
        end
        self.InRange = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams)
            local boolean = self.OvaleSpells:IsSpellInRange(spellId, target)
            return ReturnBoolean(boolean or false)
        end
        self.IsAggroed = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local boolean = UnitDetailedThreatSituation("player", target)
            return ReturnBoolean(boolean or false)
        end
        self.IsDead = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local boolean = UnitIsDead(target)
            return ReturnBoolean(boolean or false)
        end
        self.IsEnraged = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, "enrage", atTime, "HELPFUL", false)
            if aura then
                local gain, _, ending = aura.gain, aura.start, aura.ending
                return gain, ending
            end
            return 
        end
        self.IsFeared = function(positionalParams, namedParams, atTime)
            local boolean =  not HasFullControl() and self.OvaleLossOfControl.HasLossOfControl("FEAR", atTime)
            return ReturnBoolean(boolean)
        end
        self.IsFriend = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local boolean = UnitIsFriend("player", target)
            return ReturnBoolean(boolean)
        end
        self.IsIncapacitated = function(positionalParams, namedParams, atTime)
            local boolean =  not HasFullControl() and self.OvaleLossOfControl.HasLossOfControl("CONFUSE", atTime)
            return ReturnBoolean(boolean)
        end
        self.IsInterruptible = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local name, _, _, _, _, _, _, notInterruptible = UnitCastingInfo(target)
            if  not name then
                name, _, _, _, _, _, notInterruptible = UnitChannelInfo(target)
            end
            local boolean = notInterruptible ~= nil and  not notInterruptible
            return ReturnBoolean(boolean)
        end
        self.IsPVP = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local boolean = UnitIsPVP(target)
            return ReturnBoolean(boolean)
        end
        self.IsRooted = function(positionalParams, namedParams, atTime)
            local boolean = self.OvaleLossOfControl.HasLossOfControl("ROOT", atTime)
            return ReturnBoolean(boolean)
        end
        self.IsStunned = function(positionalParams, namedParams, atTime)
            local boolean =  not HasFullControl() and self.OvaleLossOfControl.HasLossOfControl("STUN_MECHANIC", atTime)
            return ReturnBoolean(boolean)
        end
        self.ItemCharges = function(positionalParams, namedParams, atTime)
            local itemId = positionalParams[1]
            local value = GetItemCount(itemId, false, true)
            return ReturnConstant(value)
        end
        self.ItemCount = function(positionalParams, namedParams, atTime)
            local itemId = positionalParams[1]
            local value = GetItemCount(itemId)
            return ReturnConstant(value)
        end
        self.LastDamage = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local value = self.OvaleSpellDamage:Get(spellId)
            if value then
                return ReturnConstant(value)
            end
            return 
        end
        self.Level = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local value
            if target == "player" then
                value = self.OvalePaperDoll.level
            else
                value = UnitLevel(target)
            end
            return ReturnConstant(value)
        end
        self.List = function(positionalParams, namedParams, atTime)
            local name, value = positionalParams[1], positionalParams[2]
            if name and self.OvaleFrameModule.frame and self.OvaleFrameModule.frame:GetListValue(name) == value then
                return 0, INFINITY
            end
            return 
        end
        self.Name = function(atTime, target)
            return ReturnConstant(UnitName(target))
        end
        self.PTR = function(positionalParams, namedParams, atTime)
            local version, _, _, uiVersion = GetBuildInfo()
            local value = ((version > "9.0.2" or uiVersion > 90002) and 1) or 0
            return ReturnConstant(value)
        end
        self.PersistentMultiplier = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            local targetGuid = self.OvaleGUID:UnitGUID(target)
            if  not targetGuid then
                return 
            end
            local value = self.OvaleFuture:GetDamageMultiplier(spellId, targetGuid, atTime)
            return ReturnConstant(value)
        end
        self.PetPresent = function(positionalParams, namedParams, atTime)
            local name = namedParams.name
            local target = "pet"
            local boolean = UnitExists(target) and  not UnitIsDead(target) and (name == nil or name == UnitName(target))
            return ReturnBoolean(boolean)
        end
        self.AlternatePower = function(positionalParams, namedParams, atTime)
            return self:Power("alternate", positionalParams, namedParams, atTime)
        end
        self.AstralPower = function(positionalParams, namedParams, atTime)
            return self:Power("lunarpower", positionalParams, namedParams, atTime)
        end
        self.Chi = function(positionalParams, namedParams, atTime)
            return self:Power("chi", positionalParams, namedParams, atTime)
        end
        self.ComboPoints = function(positionalParams, namedParams, atTime)
            return self:Power("combopoints", positionalParams, namedParams, atTime)
        end
        self.Energy = function(positionalParams, namedParams, atTime)
            return self:Power("energy", positionalParams, namedParams, atTime)
        end
        self.Focus = function(positionalParams, namedParams, atTime)
            return self:Power("focus", positionalParams, namedParams, atTime)
        end
        self.Fury = function(positionalParams, namedParams, atTime)
            return self:Power("fury", positionalParams, namedParams, atTime)
        end
        self.HolyPower = function(positionalParams, namedParams, atTime)
            return self:Power("holypower", positionalParams, namedParams, atTime)
        end
        self.Insanity = function(positionalParams, namedParams, atTime)
            return self:Power("insanity", positionalParams, namedParams, atTime)
        end
        self.Mana = function(positionalParams, namedParams, atTime)
            return self:Power("mana", positionalParams, namedParams, atTime)
        end
        self.Maelstrom = function(positionalParams, namedParams, atTime)
            return self:Power("maelstrom", positionalParams, namedParams, atTime)
        end
        self.Pain = function(positionalParams, namedParams, atTime)
            return self:Power("pain", positionalParams, namedParams, atTime)
        end
        self.Rage = function(positionalParams, namedParams, atTime)
            return self:Power("rage", positionalParams, namedParams, atTime)
        end
        self.RunicPower = function(positionalParams, namedParams, atTime)
            return self:Power("runicpower", positionalParams, namedParams, atTime)
        end
        self.SoulShards = function(positionalParams, namedParams, atTime)
            return self:Power("soulshards", positionalParams, namedParams, atTime)
        end
        self.ArcaneCharges = function(positionalParams, namedParams, atTime)
            return self:Power("arcanecharges", positionalParams, namedParams, atTime)
        end
        self.AlternatePowerDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("alternate", positionalParams, namedParams, atTime)
        end
        self.AstralPowerDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("lunarpower", positionalParams, namedParams, atTime)
        end
        self.ChiDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("chi", positionalParams, namedParams, atTime)
        end
        self.ComboPointsDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("combopoints", positionalParams, namedParams, atTime)
        end
        self.EnergyDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("energy", positionalParams, namedParams, atTime)
        end
        self.FocusDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("focus", positionalParams, namedParams, atTime)
        end
        self.FuryDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("fury", positionalParams, namedParams, atTime)
        end
        self.HolyPowerDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("holypower", positionalParams, namedParams, atTime)
        end
        self.ManaDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("mana", positionalParams, namedParams, atTime)
        end
        self.PainDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("pain", positionalParams, namedParams, atTime)
        end
        self.RageDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("rage", positionalParams, namedParams, atTime)
        end
        self.RunicPowerDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("runicpower", positionalParams, namedParams, atTime)
        end
        self.SoulShardsDeficit = function(positionalParams, namedParams, atTime)
            return self:PowerDeficit("soulshards", positionalParams, namedParams, atTime)
        end
        self.ManaPercent = function(positionalParams, namedParams, atTime)
            return self:PowerPercent("mana", positionalParams, namedParams, atTime)
        end
        self.MaxAlternatePower = function(positionalParams, namedParams, atTime)
            return self:MaxPower("alternate", positionalParams, namedParams, atTime)
        end
        self.MaxChi = function(positionalParams, namedParams, atTime)
            return self:MaxPower("chi", positionalParams, namedParams, atTime)
        end
        self.MaxComboPoints = function(positionalParams, namedParams, atTime)
            return self:MaxPower("combopoints", positionalParams, namedParams, atTime)
        end
        self.MaxEnergy = function(positionalParams, namedParams, atTime)
            return self:MaxPower("energy", positionalParams, namedParams, atTime)
        end
        self.MaxFocus = function(positionalParams, namedParams, atTime)
            return self:MaxPower("focus", positionalParams, namedParams, atTime)
        end
        self.MaxFury = function(positionalParams, namedParams, atTime)
            return self:MaxPower("fury", positionalParams, namedParams, atTime)
        end
        self.MaxHolyPower = function(positionalParams, namedParams, atTime)
            return self:MaxPower("holypower", positionalParams, namedParams, atTime)
        end
        self.MaxMana = function(positionalParams, namedParams, atTime)
            return self:MaxPower("mana", positionalParams, namedParams, atTime)
        end
        self.MaxPain = function(positionalParams, namedParams, atTime)
            return self:MaxPower("pain", positionalParams, namedParams, atTime)
        end
        self.MaxRage = function(positionalParams, namedParams, atTime)
            return self:MaxPower("rage", positionalParams, namedParams, atTime)
        end
        self.MaxRunicPower = function(positionalParams, namedParams, atTime)
            return self:MaxPower("runicpower", positionalParams, namedParams, atTime)
        end
        self.MaxSoulShards = function(positionalParams, namedParams, atTime)
            return self:MaxPower("soulshards", positionalParams, namedParams, atTime)
        end
        self.MaxArcaneCharges = function(positionalParams, namedParams, atTime)
            return self:MaxPower("arcanecharges", positionalParams, namedParams, atTime)
        end
        self.EnergyCost = function(positionalParams, namedParams, atTime)
            return self:PowerCost("energy", positionalParams, namedParams, atTime)
        end
        self.FocusCost = function(positionalParams, namedParams, atTime)
            return self:PowerCost("focus", positionalParams, namedParams, atTime)
        end
        self.ManaCost = function(positionalParams, namedParams, atTime)
            return self:PowerCost("mana", positionalParams, namedParams, atTime)
        end
        self.RageCost = function(positionalParams, namedParams, atTime)
            return self:PowerCost("rage", positionalParams, namedParams, atTime)
        end
        self.RunicPowerCost = function(positionalParams, namedParams, atTime)
            return self:PowerCost("runicpower", positionalParams, namedParams, atTime)
        end
        self.AstralPowerCost = function(positionalParams, namedParams, atTime)
            return self:PowerCost("lunarpower", positionalParams, namedParams, atTime)
        end
        self.MainPowerCost = function(positionalParams, namedParams, atTime)
            return self:PowerCost(self.OvalePower.current.powerType, positionalParams, namedParams, atTime)
        end
        self.Present = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local boolean = UnitExists(target) and  not UnitIsDead(target)
            return ReturnBoolean(boolean)
        end
        self.PreviousGCDSpell = function(positionalParams, namedParams, atTime)
            local spell = positionalParams[1]
            local spellId = self.OvaleSpellBook:getKnownSpellId(spell)
            local count = namedParams.count
            local boolean
            if count and count > 1 then
                boolean = spellId == self.OvaleFuture.next.lastGCDSpellIds[#self.OvaleFuture.next.lastGCDSpellIds - count + 2]
            else
                boolean = spellId == self.OvaleFuture.next.lastGCDSpellId
            end
            return ReturnBoolean(boolean)
        end
        self.PreviousOffGCDSpell = function(positionalParams, namedParams, atTime)
            local spell = positionalParams[1]
            local spellId = self.OvaleSpellBook:getKnownSpellId(spell)
            local boolean = spellId == self.OvaleFuture.next.lastOffGCDSpellcast.spellId
            return ReturnBoolean(boolean)
        end
        self.PreviousSpell = function(positionalParams, namedParams, atTime)
            local spell = positionalParams[1]
            local spellId = self.OvaleSpellBook:getKnownSpellId(spell)
            local boolean = spellId == self.OvaleFuture.next.lastGCDSpellId
            return ReturnBoolean(boolean)
        end
        self.RelativeLevel = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local value, level
            if target == "player" then
                level = self.OvalePaperDoll.level
            else
                level = UnitLevel(target)
            end
            if level < 0 then
                value = 3
            else
                value = level - self.OvalePaperDoll.level
            end
            return ReturnConstant(value)
        end
        self.Refreshable = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura then
                local baseDuration = self.OvaleAura:GetBaseDuration(auraId)
                if baseDuration == INFINITY then
                    baseDuration = aura.ending - aura.start
                end
                local extensionDuration = 0.3 * baseDuration
                return aura.ending - extensionDuration, INFINITY
            end
            return 0, INFINITY
        end
        self.RemainingCastTime = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local _, _, _, startTime, endTime = UnitCastingInfo(target)
            if startTime and endTime then
                startTime = startTime / 1000
                endTime = endTime / 1000
                return ReturnValueBetween(startTime, endTime, 0, endTime, -1)
            end
            return 
        end
        self.Rune = function(positionalParams, namedParams, atTime)
            local count, startCooldown, endCooldown = self.OvaleRunes:RuneCount(atTime)
            if startCooldown < INFINITY then
                local rate = 1 / (endCooldown - startCooldown)
                return ReturnValueBetween(startCooldown, INFINITY, count, startCooldown, rate)
            end
            return ReturnConstant(count)
        end
        self.RuneDeficit = function(positionalParams, namedParams, atTime)
            local count, startCooldown, endCooldown = self.OvaleRunes:RuneDeficit(atTime)
            if startCooldown < INFINITY then
                local rate = -1 / (endCooldown - startCooldown)
                return ReturnValueBetween(startCooldown, INFINITY, count, startCooldown, rate)
            end
            return ReturnConstant(count)
        end
        self.RuneCount = function(positionalParams, namedParams, atTime)
            local count, startCooldown, endCooldown = self.OvaleRunes:RuneCount(atTime)
            if startCooldown < INFINITY then
                return ReturnValueBetween(startCooldown, endCooldown, count, startCooldown, 0)
            end
            return ReturnConstant(count)
        end
        self.TimeToRunes = function(positionalParams, namedParams, atTime)
            local runes = positionalParams[1]
            local seconds = self.OvaleRunes:GetRunesCooldown(atTime, runes)
            if seconds < 0 then
                seconds = 0
            end
            return ReturnConstant(seconds)
        end
        self.Agility = function(positionalParams, namedParams, atTime)
            return self:Snapshot("agility", 0, positionalParams, namedParams, atTime)
        end
        self.AttackPower = function(positionalParams, namedParams, atTime)
            return self:Snapshot("attackPower", 0, positionalParams, namedParams, atTime)
        end
        self.CritRating = function(positionalParams, namedParams, atTime)
            return self:Snapshot("critRating", 0, positionalParams, namedParams, atTime)
        end
        self.HasteRating = function(positionalParams, namedParams, atTime)
            return self:Snapshot("hasteRating", 0, positionalParams, namedParams, atTime)
        end
        self.Intellect = function(positionalParams, namedParams, atTime)
            return self:Snapshot("intellect", 0, positionalParams, namedParams, atTime)
        end
        self.MasteryEffect = function(positionalParams, namedParams, atTime)
            return self:Snapshot("masteryEffect", 0, positionalParams, namedParams, atTime)
        end
        self.MasteryRating = function(positionalParams, namedParams, atTime)
            return self:Snapshot("masteryRating", 0, positionalParams, namedParams, atTime)
        end
        self.MeleeCritChance = function(positionalParams, namedParams, atTime)
            return self:SnapshotCritChance("meleeCrit", 0, positionalParams, namedParams, atTime)
        end
        self.MeleeAttackSpeedPercent = function(positionalParams, namedParams, atTime)
            return self:Snapshot("meleeAttackSpeedPercent", 0, positionalParams, namedParams, atTime)
        end
        self.RangedCritChance = function(positionalParams, namedParams, atTime)
            return self:SnapshotCritChance("rangedCrit", 0, positionalParams, namedParams, atTime)
        end
        self.SpellCritChance = function(positionalParams, namedParams, atTime)
            return self:SnapshotCritChance("spellCrit", 0, positionalParams, namedParams, atTime)
        end
        self.SpellCastSpeedPercent = function(positionalParams, namedParams, atTime)
            return self:Snapshot("spellCastSpeedPercent", 0, positionalParams, namedParams, atTime)
        end
        self.Spellpower = function(positionalParams, namedParams, atTime)
            return self:Snapshot("spellPower", 0, positionalParams, namedParams, atTime)
        end
        self.Stamina = function(positionalParams, namedParams, atTime)
            return self:Snapshot("stamina", 0, positionalParams, namedParams, atTime)
        end
        self.Strength = function(positionalParams, namedParams, atTime)
            return self:Snapshot("strength", 0, positionalParams, namedParams, atTime)
        end
        self.Versatility = function(positionalParams, namedParams, atTime)
            return self:Snapshot("versatility", 0, positionalParams, namedParams, atTime)
        end
        self.VersatilityRating = function(positionalParams, namedParams, atTime)
            return self:Snapshot("versatilityRating", 0, positionalParams, namedParams, atTime)
        end
        self.Speed = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local value = (GetUnitSpeed(target) * 100) / 7
            return ReturnConstant(value)
        end
        self.SpellChargeCooldown = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local charges, maxCharges, start, duration = self.OvaleCooldown:GetSpellCharges(spellId, atTime)
            if charges and charges < maxCharges then
                local ending = start + duration
                return ReturnValueBetween(start, ending, duration, start, -1)
            end
            return ReturnConstant(0)
        end
        self.SpellCharges = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local charges, maxCharges, start, duration = self.OvaleCooldown:GetSpellCharges(spellId, atTime)
            if namedParams.count == 0 and charges < maxCharges then
                return ReturnValueBetween(atTime, INFINITY, charges + 1, start + duration, 1 / duration)
            end
            return ReturnConstant(charges)
        end
        self.SpellFullRecharge = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local charges, maxCharges, start, chargeDuration = self.OvaleCooldown:GetSpellCharges(spellId, atTime)
            if charges and charges < maxCharges then
                local duration = (maxCharges - charges) * chargeDuration
                local ending = start + duration
                return ReturnValueBetween(start, ending, duration, start, -1)
            end
            return ReturnConstant(0)
        end
        self.SpellCooldown = function(positionalParams, namedParams, atTime)
            local usable = namedParams.usable == 1
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            local targetGuid = self.OvaleGUID:UnitGUID(target)
            if  not targetGuid then
                return 
            end
            local earliest = INFINITY
            for _, spellId in ipairs(positionalParams) do
                if  not usable or self.OvaleSpells:IsUsableSpell(spellId, atTime, targetGuid) then
                    local start, duration = self.OvaleCooldown:GetSpellCooldown(spellId, atTime)
                    local t = 0
                    if start > 0 and duration > 0 then
                        t = start + duration
                    end
                    if earliest > t then
                        earliest = t
                    end
                end
            end
            if earliest == INFINITY then
                return ReturnConstant(0)
            elseif earliest > 0 then
                return ReturnValue(0, earliest, -1)
            end
            return ReturnConstant(0)
        end
        self.SpellCooldownDuration = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            local duration = self.OvaleCooldown:GetSpellCooldownDuration(spellId, atTime, target)
            return ReturnConstant(duration)
        end
        self.SpellRechargeDuration = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            local cd = self.OvaleCooldown:GetCD(spellId, atTime)
            local duration = cd.chargeDuration or self.OvaleCooldown:GetSpellCooldownDuration(spellId, atTime, target)
            return ReturnConstant(duration)
        end
        self.SpellData = function(positionalParams, namedParams, atTime)
            local spellId, key = positionalParams[1], positionalParams[2]
            local si = self.OvaleData.spellInfo[spellId]
            if si then
                local value = si[key]
                if value then
                    return ReturnConstant(value)
                end
            end
            return 
        end
        self.SpellInfoProperty = function(positionalParams, namedParams, atTime)
            local spellId, key = positionalParams[1], positionalParams[2]
            local value = self.OvaleData:GetSpellInfoProperty(spellId, atTime, key, nil)
            if value then
                return ReturnConstant(value)
            end
            return 
        end
        self.SpellCount = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local spellCount = self.OvaleSpells:GetSpellCount(spellId)
            return ReturnConstant(spellCount)
        end
        self.SpellKnown = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local boolean = self.OvaleSpellBook:IsKnownSpell(spellId)
            return ReturnBoolean(boolean)
        end
        self.SpellMaxCharges = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local _, maxCharges, _ = self.OvaleCooldown:GetSpellCharges(spellId, atTime)
            if  not maxCharges then
                return 
            end
            maxCharges = maxCharges or 1
            return ReturnConstant(maxCharges)
        end
        self.SpellUsable = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            local targetGuid = self.OvaleGUID:UnitGUID(target)
            if  not targetGuid then
                return 
            end
            local isUsable, noMana = self.OvaleSpells:IsUsableSpell(spellId, atTime, targetGuid)
            local boolean = isUsable or noMana
            return ReturnBoolean(boolean)
        end
        self.Stealthed = function(positionalParams, namedParams, atTime)
            local boolean = self.OvaleAura:GetAura("player", "stealthed_buff", atTime, "HELPFUL") ~= nil or IsStealthed()
            return ReturnBoolean(boolean)
        end
        self.LastSwing = function(positionalParams, namedParams, atTime)
            local start = 0
            OneTimeMessage("Warning: 'LastSwing()' is not implemented.")
            return ReturnValueBetween(start, INFINITY, 0, start, 1)
        end
        self.NextSwing = function(positionalParams, namedParams, atTime)
            local ending = 0
            OneTimeMessage("Warning: 'NextSwing()' is not implemented.")
            return ReturnValueBetween(0, ending, 0, ending, -1)
        end
        self.Talent = function(positionalParams, namedParams, atTime)
            local talentId = positionalParams[1]
            local boolean = self.OvaleSpellBook:GetTalentPoints(talentId) > 0
            return ReturnBoolean(boolean)
        end
        self.TalentPoints = function(positionalParams, namedParams, atTime)
            local talent = positionalParams[1]
            local value = self.OvaleSpellBook:GetTalentPoints(talent)
            return ReturnConstant(value)
        end
        self.TargetIsPlayer = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            local boolean = UnitIsUnit("player", target .. "target")
            return ReturnBoolean(boolean)
        end
        self.Threat = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams, "target")
            local _, _, value = UnitDetailedThreatSituation("player", target)
            return ReturnConstant(value)
        end
        self.TickTime = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            local tickTime
            if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                tickTime = aura.tick
            else
                tickTime = self.OvaleAura:GetTickLength(auraId, self.OvalePaperDoll.next)
            end
            if tickTime and tickTime > 0 then
                return ReturnConstant(tickTime)
            end
            return ReturnConstant(INFINITY)
        end
        self.CurrentTickTime = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            local tickTime
            if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                tickTime = aura.tick or 0
            else
                tickTime = 0
            end
            return ReturnConstant(tickTime)
        end
        self.TicksRemaining = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura then
                local tick = aura.tick
                if tick and tick > 0 then
                    return ReturnValueBetween(aura.gain, INFINITY, 1, aura.ending, -1 / tick)
                end
            end
            return ReturnConstant(0)
        end
        self.TickTimeRemaining = function(positionalParams, namedParams, atTime)
            local auraId = positionalParams[1]
            local target, filter, mine = self:ParseCondition(positionalParams, namedParams)
            local aura = self.OvaleAura:GetAura(target, auraId, atTime, filter, mine)
            if aura and self.OvaleAura:IsActiveAura(aura, atTime) then
                local lastTickTime = aura.lastTickTime or aura.start
                local tick = aura.tick or self.OvaleAura:GetTickLength(auraId, self.OvalePaperDoll.next)
                local remainingTime = tick - (atTime - lastTickTime)
                if remainingTime and remainingTime > 0 then
                    return ReturnConstant(remainingTime)
                end
            end
            return ReturnConstant(0)
        end
        self.TimeSincePreviousSpell = function(positionalParams, namedParams, atTime)
            local spell = positionalParams[1]
            local spellId = self.OvaleSpellBook:getKnownSpellId(spell)
            if  not spellId then
                return 
            end
            local t = self.OvaleFuture:TimeOfLastCast(spellId, atTime)
            return ReturnValue(0, t, 1)
        end
        self.TimeToBloodlust = function(positionalParams, namedParams, atTime)
            local value = 3600
            return ReturnConstant(value)
        end
        self.TimeToEclipse = function(positionalParams, namedParams, atTime)
            local value = 3600 * 24 * 7
            OneTimeMessage("Warning: 'TimeToEclipse()' is not implemented.")
            return ReturnValue(value, atTime, -1)
        end
        self.TimeToEnergy = function(positionalParams, namedParams, atTime)
            local level = positionalParams[1]
            return self:TimeToPower("energy", level, atTime)
        end
        self.TimeToMaxEnergy = function(positionalParams, namedParams, atTime)
            local powerType = "energy"
            local level = self.OvalePower.current.maxPower[powerType] or 0
            return self:TimeToPower(powerType, level, atTime)
        end
        self.TimeToFocus = function(positionalParams, namedParams, atTime)
            local level = positionalParams[1]
            return self:TimeToPower("focus", level, atTime)
        end
        self.TimeToMaxFocus = function(positionalParams, namedParams, atTime)
            local powerType = "focus"
            local level = self.OvalePower.current.maxPower[powerType] or 0
            return self:TimeToPower(powerType, level, atTime)
        end
        self.TimeToMaxMana = function(positionalParams, namedParams, atTime)
            local powerType = "mana"
            local level = self.OvalePower.current.maxPower[powerType] or 0
            return self:TimeToPower(powerType, level, atTime)
        end
        self.TimeToEnergyFor = function(positionalParams, namedParams, atTime)
            return self:TimeToPowerFor("energy", positionalParams, namedParams, atTime)
        end
        self.TimeToFocusFor = function(positionalParams, namedParams, atTime)
            return self:TimeToPowerFor("focus", positionalParams, namedParams, atTime)
        end
        self.TimeToSpell = function(positionalParams, namedParams, atTime)
            OneTimeMessage("Warning: 'TimeToSpell()' is not implemented.")
            return ReturnValue(0, atTime, -1)
        end
        self.TimeWithHaste = function(positionalParams, namedParams, atTime)
            local seconds = positionalParams[1]
            local haste = (namedParams.haste) or "spell"
            local value = self:GetHastedTime(seconds, haste)
            return ReturnConstant(value)
        end
        self.TotemExpires = function(positionalParams, namedParams, atTime)
            local id = positionalParams[1]
            local seconds = positionalParams[2]
            seconds = seconds or 0
            local count, _, ending = self.OvaleTotem:GetTotemInfo(id, atTime)
            if count ~= nil and ending ~= nil and count > 0 then
                return ending - seconds, INFINITY
            end
            return 0, INFINITY
        end
        self.TotemPresent = function(positionalParams, namedParams, atTime)
            local id = positionalParams[1]
            local count, start, ending = self.OvaleTotem:GetTotemInfo(id, atTime)
            if count ~= nil and ending ~= nil and start ~= nil and count > 0 then
                return start, ending
            end
            return 
        end
        self.TotemRemaining = function(positionalParams, namedParams, atTime)
            local id = positionalParams[1]
            local count, start, ending = self.OvaleTotem:GetTotemInfo(id, atTime)
            if count ~= nil and start ~= nil and ending ~= nil and count > 0 then
                return ReturnValueBetween(start, ending, 0, ending, -1)
            end
            return ReturnConstant(0)
        end
        self.Tracking = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local spellName = self.OvaleSpellBook:GetSpellName(spellId)
            local numTrackingTypes = GetNumTrackingTypes()
            local boolean = false
            for i = 1, numTrackingTypes, 1 do
                local name, _, active = GetTrackingInfo(i)
                if name and name == spellName then
                    boolean = active == 1
                    break
                end
            end
            return ReturnBoolean(boolean)
        end
        self.TravelTime = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local si = spellId and self.OvaleData.spellInfo[spellId]
            local travelTime = 0
            if si then
                travelTime = si.travel_time or si.max_travel_time or 0
            end
            if travelTime > 0 then
                local estimatedTravelTime = 1
                if travelTime < estimatedTravelTime then
                    travelTime = estimatedTravelTime
                end
            end
            return ReturnConstant(travelTime)
        end
        self.True = function(positionalParams, namedParams, atTime)
            return 0, INFINITY
        end
        self.WeaponDPS = function(positionalParams, namedParams, atTime)
            local hand = positionalParams[1]
            local value = 0
            if hand == "offhand" or hand == "off" then
                value = self.OvalePaperDoll.current.offHandWeaponDPS or 0
            elseif hand == "mainhand" or hand == "main" then
                value = self.OvalePaperDoll.current.mainHandWeaponDPS or 0
            else
                value = self.OvalePaperDoll.current.mainHandWeaponDPS or 0
            end
            return ReturnConstant(value)
        end
        self.SigilCharging = function(positionalParams, namedParams, atTime)
            local charging = false
            for _, v in ipairs(positionalParams) do
                charging = charging or self.OvaleSigil:IsSigilCharging(v, atTime)
            end
            return ReturnBoolean(charging)
        end
        self.IsBossFight = function(positionalParams, namedParams, atTime)
            local bossEngaged = self.OvaleBossMod:IsBossEngaged(atTime)
            return ReturnBoolean(bossEngaged)
        end
        self.Race = function(positionalParams, namedParams, atTime)
            local isRace = false
            local target = (namedParams.target) or "player"
            local _, targetRaceId = UnitRace(target)
            for _, v in ipairs(positionalParams) do
                isRace = isRace or v == targetRaceId
            end
            return ReturnBoolean(isRace)
        end
        self.UnitInPartyCond = function(positionalParams, namedParams, atTime)
            local target = (namedParams.target) or "player"
            local boolean = UnitInParty(target)
            return ReturnBoolean(boolean)
        end
        self.UnitInRaidCond = function(positionalParams, namedParams, atTime)
            local target = (namedParams.target) or "player"
            local raidIndex = UnitInRaid(target)
            return ReturnBoolean(raidIndex ~= nil)
        end
        self.SoulFragments = function(positionalParams, namedParams, atTime)
            local value = self.OvaleDemonHunterSoulFragments:SoulFragments(atTime)
            return ReturnConstant(value)
        end
        self.HasDebuffType = function(positionalParams, namedParams, atTime)
            local target = self:ParseCondition(positionalParams, namedParams)
            for _, debuffType in ipairs(positionalParams) do
                local aura = self.OvaleAura:GetAura(target, lower(debuffType), atTime, (target == "player" and "HARMFUL") or "HELPFUL", false)
                if aura then
                    local gain, _, ending = aura.gain, aura.start, aura.ending
                    return gain, ending
                end
            end
            return 
        end
        self.stackTimeTo = function(positionalParams, namedParams, atTime)
            local spellId = positionalParams[1]
            local stacks = positionalParams[2]
            local direction = positionalParams[3]
            local incantersFlowBuff = self.OvaleData:GetSpellInfo(spellId)
            local tickCycle = (incantersFlowBuff and (incantersFlowBuff.max_stacks or 5) * 2) or 0
            local posLo
            local posHi
            if direction == "up" then
                posLo = stacks
                posHi = stacks
            elseif direction == "down" then
                posLo = tickCycle - stacks + 1
                posHi = posLo
            else
                posLo = stacks
                posHi = tickCycle - stacks + 1
            end
            local aura = self.OvaleAura:GetAura("player", spellId, atTime, "HELPFUL")
            if  not aura or aura.tick == nil or aura.lastTickTime == nil then
                return 
            end
            local buffPos
            local buffStacks = aura.stacks
            if aura.direction < 0 then
                buffPos = tickCycle - buffStacks + 1
            else
                buffPos = buffStacks
            end
            if posLo == buffPos or posHi == buffPos then
                return ReturnValue(0, 0, 0)
            end
            local ticksLo = (tickCycle + posLo - buffPos) % tickCycle
            local ticksHi = (tickCycle + posHi - buffPos) % tickCycle
            local tickTime = aura.tick
            local tickRem = tickTime - (atTime - aura.lastTickTime)
            local value = tickRem + tickTime * (min(ticksLo, ticksHi) - 1)
            return ReturnValue(value, atTime, -1)
        end
        self.message = function(positionalParameters)
            OneTimeMessage(positionalParameters[1])
            return ReturnConstant(0)
        end
        ovaleCondition:RegisterCondition("message", false, self.message)
        ovaleCondition:RegisterCondition("present", false, self.Present)
        ovaleCondition:RegisterCondition("stacktimeto", false, self.stackTimeTo)
        ovaleCondition:RegisterCondition("armorsetbonus", false, self.ArmorSetBonus)
        ovaleCondition:RegisterCondition("armorsetparts", false, self.ArmorSetParts)
        ovaleCondition:RegisterCondition("azeriteessenceismajor", false, self.AzeriteEssenceIsMajor)
        ovaleCondition:RegisterCondition("azeriteessenceisminor", false, self.AzeriteEssenceIsMinor)
        ovaleCondition:RegisterCondition("azeriteessenceisenabled", false, self.AzeriteEssenceIsEnabled)
        ovaleCondition:RegisterCondition("azeriteessencerank", false, self.AzeriteEssenceRank)
        ovaleCondition:RegisterCondition("baseduration", false, self.BaseDuration)
        ovaleCondition:RegisterCondition("buffdurationifapplied", false, self.BaseDuration)
        ovaleCondition:RegisterCondition("debuffdurationifapplied", false, self.BaseDuration)
        ovaleCondition:RegisterCondition("buffamount", false, self.BuffAmount)
        ovaleCondition:RegisterCondition("debuffamount", false, self.BuffAmount)
        ovaleCondition:RegisterCondition("tickvalue", false, self.BuffAmount)
        ovaleCondition:RegisterCondition("buffcombopoints", false, self.BuffComboPoints)
        ovaleCondition:RegisterCondition("debuffcombopoints", false, self.BuffComboPoints)
        ovaleCondition:RegisterCondition("buffcooldown", false, self.BuffCooldown)
        ovaleCondition:RegisterCondition("debuffcooldown", false, self.BuffCooldown)
        ovaleCondition:RegisterCondition("buffcount", false, self.BuffCount)
        ovaleCondition:RegisterCondition("buffcooldownduration", false, self.BuffCooldownDuration)
        ovaleCondition:RegisterCondition("debuffcooldownduration", false, self.BuffCooldownDuration)
        ovaleCondition:RegisterCondition("buffcountonany", false, self.BuffCountOnAny)
        ovaleCondition:RegisterCondition("debuffcountonany", false, self.BuffCountOnAny)
        ovaleCondition:RegisterCondition("buffdirection", false, self.BuffDirection)
        ovaleCondition:RegisterCondition("debuffdirection", false, self.BuffDirection)
        ovaleCondition:RegisterCondition("buffduration", false, self.BuffDuration)
        ovaleCondition:RegisterCondition("debuffduration", false, self.BuffDuration)
        ovaleCondition:RegisterCondition("buffexpires", false, self.BuffExpires)
        ovaleCondition:RegisterCondition("debuffexpires", false, self.BuffExpires)
        local targetParameter = {
            name = "target",
            optional = true,
            defaultValue = "player",
            type = "string"
        }
        local filterParameter = {
            name = "filter",
            optional = true,
            defaultValue = "buff",
            type = "string",
            mapValues = mapFilter
        }
        local mineParameter = {
            name = "mine",
            type = "boolean",
            defaultValue = true,
            optional = true
        }
        ovaleCondition:register("buffpresent", self.BuffPresent, {
            type = "number"
        }, {
            name = "aura",
            type = "number",
            isSpell = true,
            optional = false
        }, targetParameter, filterParameter, mineParameter, {
            name = "seconds",
            type = "number",
            defaultValue = 0,
            optional = true
        }, {
            name = "haste",
            type = "string",
            checkTokens = checkHaste,
            defaultValue = "none",
            optional = true
        })
        ovaleCondition:registerAlias("buffpresent", "debuffpresent")
        ovaleCondition:RegisterCondition("buffgain", false, self.BuffGain)
        ovaleCondition:RegisterCondition("debuffgain", false, self.BuffGain)
        ovaleCondition:RegisterCondition("buffimproved", false, self.BuffImproved)
        ovaleCondition:RegisterCondition("debuffimproved", false, self.BuffImproved)
        ovaleCondition:RegisterCondition("buffpersistentmultiplier", false, self.BuffPersistentMultiplier)
        ovaleCondition:RegisterCondition("debuffpersistentmultiplier", false, self.BuffPersistentMultiplier)
        ovaleCondition:RegisterCondition("buffremaining", false, self.BuffRemaining)
        ovaleCondition:RegisterCondition("debuffremaining", false, self.BuffRemaining)
        ovaleCondition:RegisterCondition("buffremains", false, self.BuffRemaining)
        ovaleCondition:RegisterCondition("debuffremains", false, self.BuffRemaining)
        ovaleCondition:RegisterCondition("buffremainingonany", false, self.BuffRemainingOnAny)
        ovaleCondition:RegisterCondition("debuffremainingonany", false, self.BuffRemainingOnAny)
        ovaleCondition:RegisterCondition("buffremainsonany", false, self.BuffRemainingOnAny)
        ovaleCondition:RegisterCondition("debuffremainsonany", false, self.BuffRemainingOnAny)
        ovaleCondition:RegisterCondition("buffstacks", false, self.BuffStacks)
        ovaleCondition:RegisterCondition("debuffstacks", false, self.BuffStacks)
        ovaleCondition:RegisterCondition("maxstacks", true, self.maxStacks)
        ovaleCondition:RegisterCondition("buffstacksonany", false, self.BuffStacksOnAny)
        ovaleCondition:RegisterCondition("debuffstacksonany", false, self.BuffStacksOnAny)
        ovaleCondition:RegisterCondition("buffstealable", false, self.BuffStealable)
        ovaleCondition:RegisterCondition("cancast", true, self.CanCast)
        ovaleCondition:RegisterCondition("casttime", true, self.CastTime)
        ovaleCondition:RegisterCondition("executetime", true, self.ExecuteTime)
        ovaleCondition:RegisterCondition("casting", false, self.Casting)
        ovaleCondition:RegisterCondition("checkboxoff", false, self.CheckBoxOff)
        ovaleCondition:RegisterCondition("checkboxon", false, self.CheckBoxOn)
        ovaleCondition:RegisterCondition("class", false, self.Class)
        ovaleCondition:RegisterCondition("classification", false, self.Classification)
        ovaleCondition:RegisterCondition("counter", false, self.Counter)
        ovaleCondition:register("creaturefamily", self.CreatureFamily, {
            type = "none"
        }, {
            name = "name",
            type = "string",
            optional = false
        }, targetParameter)
        ovaleCondition:RegisterCondition("creaturetype", false, self.CreatureType)
        ovaleCondition:RegisterCondition("critdamage", false, self.CritDamage)
        ovaleCondition:RegisterCondition("damage", false, self.Damage)
        ovaleCondition:RegisterCondition("damagetaken", false, self.DamageTaken)
        ovaleCondition:RegisterCondition("incomingdamage", false, self.DamageTaken)
        ovaleCondition:RegisterCondition("magicdamagetaken", false, self.MagicDamageTaken)
        ovaleCondition:RegisterCondition("incomingmagicdamage", false, self.MagicDamageTaken)
        ovaleCondition:RegisterCondition("physicaldamagetaken", false, self.PhysicalDamageTaken)
        ovaleCondition:RegisterCondition("incomingphysicaldamage", false, self.PhysicalDamageTaken)
        ovaleCondition:RegisterCondition("diseasesremaining", false, self.DiseasesRemaining)
        ovaleCondition:RegisterCondition("diseasesticking", false, self.DiseasesTicking)
        ovaleCondition:RegisterCondition("diseasesanyticking", false, self.DiseasesAnyTicking)
        ovaleCondition:RegisterCondition("distance", false, self.Distance)
        ovaleCondition:RegisterCondition("enemies", false, self.Enemies)
        ovaleCondition:RegisterCondition("energyregen", false, self.EnergyRegenRate)
        ovaleCondition:RegisterCondition("energyregenrate", false, self.EnergyRegenRate)
        ovaleCondition:RegisterCondition("enrageremaining", false, self.EnrageRemaining)
        ovaleCondition:RegisterCondition("exists", false, self.Exists)
        ovaleCondition:RegisterCondition("never", false, self.False)
        ovaleCondition:RegisterCondition("focusregen", false, self.FocusRegenRate)
        ovaleCondition:RegisterCondition("focusregenrate", false, self.FocusRegenRate)
        ovaleCondition:RegisterCondition("focuscastingregen", false, self.FocusCastingRegen)
        ovaleCondition:RegisterCondition("gcd", false, self.GCD)
        ovaleCondition:RegisterCondition("gcdremaining", false, self.GCDRemaining)
        ovaleCondition:RegisterCondition("glyph", false, self.Glyph)
        ovaleCondition:RegisterCondition("hasfullcontrol", false, self.HasFullControlCondition)
        ovaleCondition:RegisterCondition("health", false, self.Health)
        ovaleCondition:RegisterCondition("life", false, self.Health)
        ovaleCondition:RegisterCondition("effectivehealth", false, self.EffectiveHealth)
        ovaleCondition:RegisterCondition("healthmissing", false, self.HealthMissing)
        ovaleCondition:RegisterCondition("lifemissing", false, self.HealthMissing)
        ovaleCondition:RegisterCondition("healthpercent", false, self.HealthPercent)
        ovaleCondition:RegisterCondition("lifepercent", false, self.HealthPercent)
        ovaleCondition:RegisterCondition("effectivehealthpercent", false, self.EffectiveHealthPercent)
        ovaleCondition:RegisterCondition("maxhealth", false, self.MaxHealth)
        ovaleCondition:RegisterCondition("deadin", false, self.TimeToDie)
        ovaleCondition:RegisterCondition("timetodie", false, self.TimeToDie)
        ovaleCondition:RegisterCondition("timetohealthpercent", false, self.TimeToHealthPercent)
        ovaleCondition:RegisterCondition("timetolifepercent", false, self.TimeToHealthPercent)
        ovaleCondition:RegisterCondition("inflighttotarget", false, self.InFlightToTarget)
        ovaleCondition:RegisterCondition("inrange", false, self.InRange)
        ovaleCondition:RegisterCondition("isaggroed", false, self.IsAggroed)
        ovaleCondition:RegisterCondition("isdead", false, self.IsDead)
        ovaleCondition:RegisterCondition("isenraged", false, self.IsEnraged)
        ovaleCondition:RegisterCondition("isfeared", false, self.IsFeared)
        ovaleCondition:RegisterCondition("isfriend", false, self.IsFriend)
        ovaleCondition:RegisterCondition("isincapacitated", false, self.IsIncapacitated)
        ovaleCondition:RegisterCondition("isinterruptible", false, self.IsInterruptible)
        ovaleCondition:RegisterCondition("ispvp", false, self.IsPVP)
        ovaleCondition:RegisterCondition("isrooted", false, self.IsRooted)
        ovaleCondition:RegisterCondition("isstunned", false, self.IsStunned)
        ovaleCondition:RegisterCondition("itemcharges", false, self.ItemCharges)
        ovaleCondition:RegisterCondition("itemcount", false, self.ItemCount)
        ovaleCondition:RegisterCondition("lastdamage", false, self.LastDamage)
        ovaleCondition:RegisterCondition("lastspelldamage", false, self.LastDamage)
        ovaleCondition:RegisterCondition("level", false, self.Level)
        ovaleCondition:RegisterCondition("list", false, self.List)
        ovaleCondition:register("name", self.Name, {
            type = "string"
        }, targetParameter)
        ovaleCondition:RegisterCondition("ptr", false, self.PTR)
        ovaleCondition:RegisterCondition("persistentmultiplier", false, self.PersistentMultiplier)
        ovaleCondition:RegisterCondition("petpresent", false, self.PetPresent)
        ovaleCondition:RegisterCondition("alternatepower", false, self.AlternatePower)
        ovaleCondition:RegisterCondition("arcanecharges", false, self.ArcaneCharges)
        ovaleCondition:RegisterCondition("astralpower", false, self.AstralPower)
        ovaleCondition:RegisterCondition("chi", false, self.Chi)
        ovaleCondition:RegisterCondition("combopoints", false, self.ComboPoints)
        ovaleCondition:RegisterCondition("energy", false, self.Energy)
        ovaleCondition:RegisterCondition("focus", false, self.Focus)
        ovaleCondition:RegisterCondition("fury", false, self.Fury)
        ovaleCondition:RegisterCondition("holypower", false, self.HolyPower)
        ovaleCondition:RegisterCondition("insanity", false, self.Insanity)
        ovaleCondition:RegisterCondition("maelstrom", false, self.Maelstrom)
        ovaleCondition:RegisterCondition("mana", false, self.Mana)
        ovaleCondition:RegisterCondition("pain", false, self.Pain)
        ovaleCondition:RegisterCondition("rage", false, self.Rage)
        ovaleCondition:RegisterCondition("runicpower", false, self.RunicPower)
        ovaleCondition:RegisterCondition("soulshards", false, self.SoulShards)
        ovaleCondition:RegisterCondition("alternatepowerdeficit", false, self.AlternatePowerDeficit)
        ovaleCondition:RegisterCondition("astralpowerdeficit", false, self.AstralPowerDeficit)
        ovaleCondition:RegisterCondition("chideficit", false, self.ChiDeficit)
        ovaleCondition:RegisterCondition("combopointsdeficit", false, self.ComboPointsDeficit)
        ovaleCondition:RegisterCondition("energydeficit", false, self.EnergyDeficit)
        ovaleCondition:RegisterCondition("focusdeficit", false, self.FocusDeficit)
        ovaleCondition:RegisterCondition("furydeficit", false, self.FuryDeficit)
        ovaleCondition:RegisterCondition("holypowerdeficit", false, self.HolyPowerDeficit)
        ovaleCondition:RegisterCondition("manadeficit", false, self.ManaDeficit)
        ovaleCondition:RegisterCondition("paindeficit", false, self.PainDeficit)
        ovaleCondition:RegisterCondition("ragedeficit", false, self.RageDeficit)
        ovaleCondition:RegisterCondition("runicpowerdeficit", false, self.RunicPowerDeficit)
        ovaleCondition:RegisterCondition("soulshardsdeficit", false, self.SoulShardsDeficit)
        ovaleCondition:RegisterCondition("manapercent", false, self.ManaPercent)
        ovaleCondition:RegisterCondition("maxalternatepower", false, self.MaxAlternatePower)
        ovaleCondition:RegisterCondition("maxarcanecharges", false, self.MaxArcaneCharges)
        ovaleCondition:RegisterCondition("maxchi", false, self.MaxChi)
        ovaleCondition:RegisterCondition("maxcombopoints", false, self.MaxComboPoints)
        ovaleCondition:RegisterCondition("maxenergy", false, self.MaxEnergy)
        ovaleCondition:RegisterCondition("maxfocus", false, self.MaxFocus)
        ovaleCondition:RegisterCondition("maxfury", false, self.MaxFury)
        ovaleCondition:RegisterCondition("maxholypower", false, self.MaxHolyPower)
        ovaleCondition:RegisterCondition("maxmana", false, self.MaxMana)
        ovaleCondition:RegisterCondition("maxpain", false, self.MaxPain)
        ovaleCondition:RegisterCondition("maxrage", false, self.MaxRage)
        ovaleCondition:RegisterCondition("maxrunicpower", false, self.MaxRunicPower)
        ovaleCondition:RegisterCondition("maxsoulshards", false, self.MaxSoulShards)
        ovaleCondition:RegisterCondition("powercost", true, self.MainPowerCost)
        ovaleCondition:RegisterCondition("astralpowercost", true, self.AstralPowerCost)
        ovaleCondition:RegisterCondition("energycost", true, self.EnergyCost)
        ovaleCondition:RegisterCondition("focuscost", true, self.FocusCost)
        ovaleCondition:RegisterCondition("manacost", true, self.ManaCost)
        ovaleCondition:RegisterCondition("ragecost", true, self.RageCost)
        ovaleCondition:RegisterCondition("runicpowercost", true, self.RunicPowerCost)
        ovaleCondition:RegisterCondition("previousgcdspell", true, self.PreviousGCDSpell)
        ovaleCondition:RegisterCondition("previousoffgcdspell", true, self.PreviousOffGCDSpell)
        ovaleCondition:RegisterCondition("previousspell", true, self.PreviousSpell)
        ovaleCondition:RegisterCondition("relativelevel", false, self.RelativeLevel)
        ovaleCondition:RegisterCondition("refreshable", false, self.Refreshable)
        ovaleCondition:RegisterCondition("debuffrefreshable", false, self.Refreshable)
        ovaleCondition:RegisterCondition("buffrefreshable", false, self.Refreshable)
        ovaleCondition:RegisterCondition("remainingcasttime", false, self.RemainingCastTime)
        ovaleCondition:RegisterCondition("rune", false, self.Rune)
        ovaleCondition:RegisterCondition("runecount", false, self.RuneCount)
        ovaleCondition:RegisterCondition("timetorunes", false, self.TimeToRunes)
        ovaleCondition:RegisterCondition("runedeficit", false, self.RuneDeficit)
        ovaleCondition:RegisterCondition("agility", false, self.Agility)
        ovaleCondition:RegisterCondition("attackpower", false, self.AttackPower)
        ovaleCondition:RegisterCondition("critrating", false, self.CritRating)
        ovaleCondition:RegisterCondition("hasterating", false, self.HasteRating)
        ovaleCondition:RegisterCondition("intellect", false, self.Intellect)
        ovaleCondition:RegisterCondition("mastery", false, self.MasteryEffect)
        ovaleCondition:RegisterCondition("masteryeffect", false, self.MasteryEffect)
        ovaleCondition:RegisterCondition("masteryrating", false, self.MasteryRating)
        ovaleCondition:RegisterCondition("meleecritchance", false, self.MeleeCritChance)
        ovaleCondition:RegisterCondition("meleeattackspeedpercent", false, self.MeleeAttackSpeedPercent)
        ovaleCondition:RegisterCondition("rangedcritchance", false, self.RangedCritChance)
        ovaleCondition:RegisterCondition("spellcritchance", false, self.SpellCritChance)
        ovaleCondition:RegisterCondition("spellcastspeedpercent", false, self.SpellCastSpeedPercent)
        ovaleCondition:RegisterCondition("spellpower", false, self.Spellpower)
        ovaleCondition:RegisterCondition("stamina", false, self.Stamina)
        ovaleCondition:RegisterCondition("strength", false, self.Strength)
        ovaleCondition:RegisterCondition("versatility", false, self.Versatility)
        ovaleCondition:RegisterCondition("versatilityRating", false, self.VersatilityRating)
        ovaleCondition:RegisterCondition("speed", false, self.Speed)
        ovaleCondition:RegisterCondition("spellchargecooldown", true, self.SpellChargeCooldown)
        ovaleCondition:RegisterCondition("charges", true, self.SpellCharges)
        ovaleCondition:RegisterCondition("spellcharges", true, self.SpellCharges)
        ovaleCondition:RegisterCondition("spellfullrecharge", true, self.SpellFullRecharge)
        ovaleCondition:RegisterCondition("spellcooldown", true, self.SpellCooldown)
        ovaleCondition:RegisterCondition("spellcooldownduration", true, self.SpellCooldownDuration)
        ovaleCondition:RegisterCondition("spellrechargeduration", true, self.SpellRechargeDuration)
        ovaleCondition:RegisterCondition("spelldata", false, self.SpellData)
        ovaleCondition:RegisterCondition("spellinfoproperty", false, self.SpellInfoProperty)
        ovaleCondition:RegisterCondition("spellcount", true, self.SpellCount)
        ovaleCondition:RegisterCondition("spellknown", true, self.SpellKnown)
        ovaleCondition:RegisterCondition("spellmaxcharges", true, self.SpellMaxCharges)
        ovaleCondition:RegisterCondition("spellusable", true, self.SpellUsable)
        ovaleCondition:RegisterCondition("isstealthed", false, self.Stealthed)
        ovaleCondition:RegisterCondition("stealthed", false, self.Stealthed)
        ovaleCondition:RegisterCondition("lastswing", false, self.LastSwing)
        ovaleCondition:RegisterCondition("nextswing", false, self.NextSwing)
        ovaleCondition:RegisterCondition("talent", false, self.Talent)
        ovaleCondition:RegisterCondition("hastalent", false, self.Talent)
        ovaleCondition:RegisterCondition("talentpoints", false, self.TalentPoints)
        ovaleCondition:RegisterCondition("istargetingplayer", false, self.TargetIsPlayer)
        ovaleCondition:RegisterCondition("targetisplayer", false, self.TargetIsPlayer)
        ovaleCondition:RegisterCondition("threat", false, self.Threat)
        ovaleCondition:RegisterCondition("ticktime", false, self.TickTime)
        ovaleCondition:RegisterCondition("currentticktime", false, self.CurrentTickTime)
        ovaleCondition:RegisterCondition("ticksremaining", false, self.TicksRemaining)
        ovaleCondition:RegisterCondition("ticksremain", false, self.TicksRemaining)
        ovaleCondition:RegisterCondition("ticktimeremaining", false, self.TickTimeRemaining)
        ovaleCondition:RegisterCondition("timesincepreviousspell", false, self.TimeSincePreviousSpell)
        ovaleCondition:RegisterCondition("timetobloodlust", false, self.TimeToBloodlust)
        ovaleCondition:RegisterCondition("timetoeclipse", false, self.TimeToEclipse)
        ovaleCondition:RegisterCondition("timetoenergy", false, self.TimeToEnergy)
        ovaleCondition:RegisterCondition("timetofocus", false, self.TimeToFocus)
        ovaleCondition:RegisterCondition("timetomaxenergy", false, self.TimeToMaxEnergy)
        ovaleCondition:RegisterCondition("timetomaxfocus", false, self.TimeToMaxFocus)
        ovaleCondition:RegisterCondition("timetomaxmana", false, self.TimeToMaxMana)
        ovaleCondition:RegisterCondition("timetoenergyfor", true, self.TimeToEnergyFor)
        ovaleCondition:RegisterCondition("timetofocusfor", true, self.TimeToFocusFor)
        ovaleCondition:RegisterCondition("timetospell", true, self.TimeToSpell)
        ovaleCondition:RegisterCondition("timewithhaste", false, self.TimeWithHaste)
        ovaleCondition:RegisterCondition("totemexpires", false, self.TotemExpires)
        ovaleCondition:RegisterCondition("totempresent", false, self.TotemPresent)
        ovaleCondition:RegisterCondition("totemremaining", false, self.TotemRemaining)
        ovaleCondition:RegisterCondition("totemremains", false, self.TotemRemaining)
        ovaleCondition:RegisterCondition("tracking", false, self.Tracking)
        ovaleCondition:RegisterCondition("traveltime", true, self.TravelTime)
        ovaleCondition:RegisterCondition("maxtraveltime", true, self.TravelTime)
        ovaleCondition:RegisterCondition("always", false, self.True)
        ovaleCondition:RegisterCondition("weapondps", false, self.WeaponDPS)
        ovaleCondition:RegisterCondition("sigilcharging", false, self.SigilCharging)
        ovaleCondition:RegisterCondition("isbossfight", false, self.IsBossFight)
        ovaleCondition:RegisterCondition("race", false, self.Race)
        ovaleCondition:RegisterCondition("unitinparty", false, self.UnitInPartyCond)
        ovaleCondition:RegisterCondition("unitinraid", false, self.UnitInRaidCond)
        ovaleCondition:RegisterCondition("soulfragments", false, self.SoulFragments)
        ovaleCondition:RegisterCondition("hasdebufftype", false, self.HasDebuffType)
    end,
})
